
`include "aes_basic_func.v"
`include "aes_basic_fb_con.v"

module aes_basic (
  input          CLK,
  input          RST,
  input          START,
  input  [31:00] DIN, 
  input  [31:00] KEY, 
  
  output [31:00] DOUT
);

  // state machine parameters
  parameter S0=4'd0, S1=4'd1, S2=4'd2, S3=4'd3, S4=4'd4, S5=4'd5, S6=4'd6, S7=4'd7, S8=4'd8;

  reg    [05:00] count;
  reg    [03:00] sm_state;     // state machine register
  reg    [01:00] aes_mux_sel;  // controls which feedback registers are loaded
  reg    [03:00] bram_ctrl;    // control for T-table selection
  reg    [03:00] last;         // last round selection
  reg    [03:00] tshift;       // control for shifting output of BRAM
  reg            ptext_nkey;   // chooses between plaintext input and feedback
 
  wire   [31:00] aes_din;
  wire   [31:00] aes_dout;

  assign DOUT = aes_dout;

  // the feedback controler
  aes_fb_con fbc_0 (.CLK  (CLK                      ), 
                    .CTRL ({ptext_nkey, aes_mux_sel}), 
                    .DIN0 (aes_dout                 ), 
                    .DIN1 (DIN                      ),
                 
                    .DOUT (aes_din                  )
                   );
  
  // the basic function of the AES (quarter round) 
  aes_basic_func f_0 (.CLK   (CLK              ),
                      .RST   (RST              ),
                      .KEY   (KEY              ),
                      .DIN   (aes_din          ),
                      .BCT   ({bram_ctrl, last}),
                      .TSH   (tshift           ),
               
                      .DOUT  (aes_dout         )
                     );

/* Note that signals "tshift" and "bram_ctrl" are 4 bit wide, though in 
the state machine they are set to the same value. The synthesized will 
combine these registers into a single signal; they are left in the code 
as they are for readability and also such that it will be easier to modify 
the T-table order for decryption */

  always @(posedge CLK)
  begin
    if (RST) 
    begin
      count        <= 0;
      sm_state     <= S0;
      ptext_nkey   <= 1;
      aes_mux_sel  <= 2'b00;
      bram_ctrl    <= 4'b0101;
      tshift       <= 4'b0101;
      last         <= 4'b0000;
    end  
    else
    begin

      // shift right last round control
      last[2:0]      <= last[3:1];
    
      case (sm_state)
      
        S0  :  begin
                 aes_mux_sel    <= 2'b00;  
                 tshift         <= 4'b0101;
                 bram_ctrl      <= 4'b0101;
                 ptext_nkey     <= 1'b1;
                 if (START == 1'b1)
                   sm_state     <= S1;
                 else
                   sm_state     <= S0;  
               end        
 
        S1  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b1010;
                 tshift         <= 4'b1111;
                 sm_state       <= S2;
               end

        S2  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b1010;
                 tshift         <= 4'b0000;
                 sm_state       <= S3;
               end

        S3  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b0101;
                 tshift         <= 4'b1111;
                 sm_state       <= S4;
               end        

        S4  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b0101;
                 tshift         <= 4'b0000;
                 sm_state       <= S5;
               end
                 
        S5  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b1010;
                 tshift         <= 4'b1111;
                 sm_state       <= S6;
               end  
               
        S6  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b1010;
                 tshift         <= 4'b0000;
                 sm_state       <= S7;
               end

        S7  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b0101;
                 tshift         <= 4'b1111;
                 sm_state       <= S8;
               end

        S8  :  begin
                 aes_mux_sel    <= aes_mux_sel + 1;
                 bram_ctrl      <= 4'b0101;
                 tshift         <= 4'b0000;
                 
                 count          <= count + 1; // increment round counter
                 
                 ptext_nkey     <= 1'b0;  // after first round, switch to feedback input
                                 
                 if (count == 4'd8)  // last round control
                   last[3] <= 1;
                                       
                 sm_state       <= S1;
               end
               
        default : begin
                    sm_state <= S0;
                  end
      endcase
    end             
  end

endmodule
